import java.io.*;
import java.util.*;
public class Matching{

    //PRE: bg.sizeLeft>0, Hall's theorem holds for the left side of bg
    //POST: creates a matching for all nodes on the left side of bg.
    Matching(BGraph bg){
	Assert.PRE(bg.sizeLeft()>0);
	int bgsize = bg.numNodes();
	this.bg = bg;
	arrMatchingEdge = new Edge [bgsize];
	getMatching();
    }

    //POST: RES is true iff n is matched.
    public boolean isMatched(Node n){
	Edge e = arrMatchingEdge[n.getUniqueNumber()];
	return (e!=null);
    }
	
    public boolean isMatched(State s){
	return isMatched(s.node());
    }

    //POST: RES is the edge n is adjacent to
    public Edge getMatchingEdge(Node n){
	Assert.PRE(n!=null);
	Edge e = arrMatchingEdge[n.getUniqueNumber()];
	Assert.POST(e!=null);
	return e;
    }

    
    //POST: RES is all of the nodes that are on the left side of the matching
    public Node[] getLeftNodes(){
	return bg.getLeftNodes();
    }

    public void SetMatchedEdge(Node n, Edge e){
	arrMatchingEdge[n.getUniqueNumber()] = e;
    }

    //POST: sets the matched edge for both endpoints of e to be e.
    public void setMatchedEdgeForEndpoints(Edge e){
	Node source = e.source();
	Node sink = e.sink();
	SetMatchedEdge(source,e);
	SetMatchedEdge(sink,e);
    }


    //PRE: vEdgeMChain is a chain of edges alternating between
    //unmatched and matched, starting with an unmatched node on the
    //left side and ending with an unmatched node on the right side.
    //firstNode is the node in the first edge which is the start of
    //the chain.  firsNode is not matched.
    //POST: toggles the states of all these edges so that one more
    //vertex on the left side and right side are matched 

    public void toggleEdgeStates(Edge arrEdgeMChain[], Node firstNode){

	Node currNode = firstNode;

	for(int i=0;i<arrEdgeMChain.length;i++){
	    if(i%2==0){
		//set the matched edge for both endpoints of the edge to be this edge.
		Edge currEdge = arrEdgeMChain[i];
		setMatchedEdgeForEndpoints(currEdge);
	    }
	}
    }

    public void ensureValidMChain(Edge arrEdgeMChain[], Node firstNode){

	Assert.PRE(!isMatched(firstNode));
	Node currNode = firstNode;

	//Debug.trace(arrEdgeMChain);

	for(int i=1;i<arrEdgeMChain.length;i++){
	    if(i%2==1){
		Edge currEdge = arrEdgeMChain[i];
		Node n1 = currEdge.source();
		Node n2 = currEdge.sink();
		Edge e1 = getMatchingEdge(n1);
		Edge e2 = getMatchingEdge(n2);
		Assert.POST(e1==currEdge);
		Assert.POST(e2==currEdge);
	    }
	}
    }


    //PRE: goal is a state containing an unmatched node with parent
    //states that eventually lead to the State start
    //POST: RES is a vector of the chain of edges which connect the start state to the goal state.

    private Edge[] getMChainFromGoalState(State sStart, State sGoal){
	Vector vEdges = new Vector();
	State currState = sGoal;

	while(currState.hasParent()){
	    Edge currEdge = currState.getEdgeToParent();
	    vEdges.addElement(currEdge);
	    currState = currState.getParentState();
	}

	Assert.POST(currState==sStart);
	MyVectorUtils.reverse(vEdges);

	Edge arrEdge[] = new Edge[vEdges.size()];
	for(int i=0;i<vEdges.size();i++){
	    arrEdge[i] = (Edge)vEdges.elementAt(i);
	}
	return arrEdge;
    }

    //POST: RES = vector containing state with a node that is
    //unvisited and reachable from currState.node() by an edge in the
    //matching if currState.pLeftSide() is false(if such a state
    //exitst).  If currState.pLeftSide() is true RES contains states
    //reachable from currState.node() by edges not in the matching.
    //The parent value of each of these states is n and the
    //edgetoParent is set appropriately.

    public Vector getChildStates(State sParent, boolean arrIsVisited[]){
	Node n = sParent.node();

	boolean onLeft = sParent.pLeftSide();
	Vector vChildren = new Vector();

	if(onLeft){
	    Edge arrEdges[] = n.getEdges();
	    for(int i=0;i<arrEdges.length;i++){
		Edge currEdge = arrEdges[i];
		Node nNeighbor = currEdge.getNeighbor(n);
		if(!isVisited(nNeighbor,arrIsVisited)){
		    State sChild = new State(nNeighbor,sParent,currEdge);
		    vChildren.addElement(sChild);
		}
	    }
	}
	else{
	    Assert.PRE(isMatched(n));
	    Edge eMatchingEdge = getMatchingEdge(n);
	    Node nNeighbor = eMatchingEdge.getNeighbor(n);
	    if(!isVisited(nNeighbor,arrIsVisited)){
		State sChild = new State(nNeighbor,sParent,eMatchingEdge);
		vChildren.addElement(sChild);
	    }
	}

	return vChildren;
    }

    private boolean isVisited(Node n, boolean arrIsVisited[]){
	int idN = n.getUniqueNumber();
	return arrIsVisited[idN];
    }

    private void markVisited(Vector vStatesChildren, boolean arrIsVisited[]){
	for(int i=0;i<vStatesChildren.size();i++){
	    State sCurr = (State)vStatesChildren.elementAt(i);
	    Node nCurr = sCurr.node();
	    int currId = nCurr.getUniqueNumber();
	    arrIsVisited[currId] = true;
	}
    }

    //PRE: n is an unmatched node.  The bipartite graph satisfies Hall's theorem
    //all nodes are marked as not visited.
    //POST: returns a chain of unmatched, matched, ...., edges starting
    //from nStart and ending at some node which is currently unmatched on the right side
    public Edge[] getAlternatingChain(Node nStart){
	
	boolean arrIsVisited[] = MyArrayUtils.getAllFalseArr(bg.numNodes());

	//uses DFS to try to find an alternating chain
	Stack stStates = new Stack();
	arrIsVisited[nStart.getUniqueNumber()] = true;

	State sStart = new State(nStart);
	State currState = sStart;

	Assert.PRE(!isMatched(nStart));

	boolean done = false;
	while(!done){
	    Vector vStatesChildren = getChildStates(currState, arrIsVisited);
	    markVisited(vStatesChildren,arrIsVisited);
	    MyVectorUtils.pushVectorOntoStack(vStatesChildren,stStates);

	    Assert.PRE(!stStates.isEmpty());  //true if Hall's Thm holds
	    currState = (State)stStates.pop();

	    done = !isMatched(currState);
	}
	Edge arrEdge[] = getMChainFromGoalState(sStart, currState);
	return arrEdge;
    }



    //POST: RES is true iff every node on the left side of the graph
    //is matched to some unique node on the right side
    public boolean isMatching(){
	int bgSize = bg.numNodes();
	int bgLeftSize = bg.sizeLeft();
	//int bgRightSize = bg.sizeRight();
	boolean arrIsMatched[] = MyArrayUtils.getAllFalseArr(bg.numNodes());
	boolean isMatching = true;

	for(int i=0;i<bgLeftSize;i++){
	    Node currLeft = bg.getLeftNode(i);
	    Assert.PRE(isMatched(currLeft));
	    Edge matchedEdge = getMatchingEdge(currLeft);
	    Node matchedRight = matchedEdge.getNeighbor(currLeft);
	    int rightID = matchedRight.getUniqueNumber();
	    Assert.POST(!arrIsMatched[rightID]);
	    arrIsMatched[rightID] = true;
	}
	return true;
    }

    //POST: n1 and n2 are matched.  
    private void matchNodes(Node n1, Node n2){
	Assert.PRE(n1!=null);
	Assert.PRE(n2!=null);

	int id1 = n1.getUniqueNumber();
	int id2 = n2.getUniqueNumber();
	Edge e = n1.getConnectingEdge(n2);

	arrMatchingEdge[id1] = e;
	arrMatchingEdge[id2] = e;
    }

    //POST: all nodes on right side that are bypass nodes (and hence have degree one) are matched
    public void matchBypassNodes(){
	for(int i=0;i<bg.sizeRight();i++){
	    Node currRight = bg.getRightNode(i);
	    if (currRight.isBypass()&&currRight.degree()!=0){
		Node[] arrNode = currRight.getNeighbors();
		Assert.PRE(arrNode.length==1);
		Node nhbr = arrNode[0];
		Assert.POST(nhbr!=null);
		if(!isMatched(nhbr)){
		    matchNodes(currRight,nhbr);
		}
	    }
	}
    }


    //PRE: BGraph has been created, all nodes on the left side have degree >= nodes on the right and 
    //maxdegree is >=1
    //POST: every node on the left side has one edge which matches it to a unique node on the right.
    public void getMatching(){
	Assert.PRE(bg.maxDegree()>0);
	matchBypassNodes();
	for(int i=0;i<bg.sizeLeft();i++){
	    Node currLeft = bg.getLeftNode(i);
	    if(!isMatched(currLeft)){
		Edge arrEdgeMChain[] = getAlternatingChain(currLeft);
		//now go through and toggle the matched state of every edge in this chain
		ensureValidMChain(arrEdgeMChain,currLeft);
		toggleEdgeStates(arrEdgeMChain, currLeft);
	    }
	}
	//System.out.println(this);
	Assert.POST(isMatching());
    }

    public String toString(){
	String sRes = "Matching\n";
	String sArr = MyIOUtils.arrToString(arrMatchingEdge,"\n");
	return sRes + sArr;
    }

    public static RGraph getPathsAndCycles(Matching match){
	Vector vRGEdges = new Vector();
	Edge arrMEdges[] = match.arrMatchingEdge;
	int arrMESize = arrMEdges.length;

	Node arrLeftNodes[] = match.getLeftNodes();
	for(int i=0;i<arrLeftNodes.length;i++){
	    int id = arrLeftNodes[i].getUniqueNumber();
	    Edge eMatch = arrMEdges[id];
	    Assert.PRE(eMatch!=null);
	    RGEdge currRGE = new RGEdge(eMatch);
	    vRGEdges.addElement(currRGE);
	}
	RGraph rgRes = new RGraph(vRGEdges);
	return rgRes;
    }


    //for each node on the left hand side contains null if the node is
    //not matched or the matching edge if it is matched
    Edge arrMatchingEdge[];
    BGraph bg;
};
